#!/usr/local/bin/perl
use strict;
use warnings;
use IO::Interactive qw(is_interactive);
use IO::Prompt;
use LWP;
use DBI;
use DBD::Pg;
use DBD::SQLite;
use RefSeq qw(@REF_SEQ_UNC);


# Edit these locations.
my $prefix = '/Users/mcb/Docs/Projects/Compression/';
my $GUNZIP = '/usr/bin/gunzip';
my $GZIP = '/usr/bin/gzip';


# Other locations - leave them alone
my $analysis_dir = $prefix . 'analysis/';
my $data_dir = $prefix . 'data/';
my $sqlite_dir = $prefix . 'sqlite/';
my $count_dir = $prefix . 'count/';
my $encoding_dir = $prefix . 'encoding/';
my $frequency_dir = $prefix . 'frequency/';
my $reference_dir = $prefix . 'reference/';


# Allow a few unix-style commands for automating with cron
if (@ARGV){#FOLDBEG
    my $option = shift;

    if ($option eq '-a') {
        #Use this area to run batch jobs from the command line.  Currently it
        #goes through all of the steps for processing the data and creating
        #encodings for chromosome 1.

        #Extract the data and generate frequencies
        #extract_hapmap_genotype_data();
        #compute_hapmap_runlength_frequencies(1);
        #compute_hapmap_variant_frequencies(1);

        #Create the encodings
        my $runlength_counts = get_key_data('1', 'runlength', 'counts');
        my $variant_counts = get_key_data('1', 'variant', 'counts');
        my $runlength_freqs = get_key_data('1', 'runlength', 'freqs');
        my $variant_freqs = get_key_data('1', 'variant', 'freqs');

        my $runlength_encoding;
        my $variant_encoding;
 
        
        #$runlength_encoding = make_encoding_elias_gamma($runlength_counts);       
        #$variant_encoding = make_encoding_elias_gamma($variant_counts);       
        #store_encoding('hapmap', 'Elias-Gamma', $runlength_encoding, 'runlength');
        #store_encoding('hapmap', 'Elias-Gamma', $variant_encoding, 'variant');

        
        #$runlength_encoding = make_encoding_golomb($runlength_counts, $runlength_freqs);
        #$variant_encoding = make_encoding_golomb($variant_counts, $variant_freqs);
        #store_encoding('hapmap', 'Golomb', $runlength_encoding, 'runlength');
        #store_encoding('hapmap', 'Golomb', $variant_encoding, 'variant');


        #$runlength_encoding = make_encoding_huffman($runlength_counts);
        #$variant_encoding = make_encoding_huffman($variant_counts);
        #store_encoding('hapmap', 'Huffman', $runlength_encoding, 'runlength');
        #store_encoding('hapmap', 'Huffman', $variant_encoding, 'variant');


        $runlength_encoding = make_encoding_unary($runlength_counts);
        $variant_encoding = make_encoding_unary($variant_counts);
        store_encoding('hapmap', 'Unary', $runlength_encoding, 'runlength');
        store_encoding('hapmap', 'Unary', $variant_encoding, 'variant');



        #Analyze the position-variant data
        analyze_encoding_hapmap_pv();


        #Analyze the encodings
        analyze_encoding( 'hapmap', 1, 'Elias-Gamma');
        analyze_encoding( 'hapmap', 1, 'Golomb');
        analyze_encoding( 'hapmap', 1, 'Huffman');
        analyze_encoding( 'hapmap', 1, 'Unary');

    }

    exit;
}#FOLDEND


# Show a menu
die "dna_compressor must be run from a terminal window"
    unless (is_interactive);

menu_top();




sub analyze_compression_factors {#FOLDBEG
    my ($data_set, $chromosome) = @_;


    # Get some numbers on the data file sizes
    my $data_file = $analysis_dir;
    if ($chromosome) {
        $data_file .= 'hapmap_chr' . $chromosome;
    }
    elsif ( $data_set =~ /mtdna/ ) {
        $data_file .= 'mtdna';
    }
    else {
        $data_file .= 'your_data'
    }
    $data_file .= '_file_sizes.txt';

    open my $INFILE, '<', $data_file
        || die "Could not open file";

    my $file_size;
    while ( <$INFILE> ) {
        chomp;
        next unless /^READ:(.+)/;
        $file_size = $1;
    }
    close $INFILE;


    # Get some numbers on the pos-value file size
    my $total_seqs;    #total number of donors for a given chromosome
    my $avg_variants;  #avg. number of variants in a seq. after consensus adjustments
    my $pv_size;       #total size(bits) for all seqs for a chromosome represented using pos-variant notation
    my $pv_file = $analysis_dir;
    if ($chromosome) {
        $pv_file .= 'hapmap_chr' . $chromosome;
    }
    else {
        $pv_file .= $data_set;
    }
    $pv_file .= '_pv.txt';

    open $INFILE, '<', $pv_file
        || die "Could not open file";
    while ( <$INFILE> ){
        chomp;
        next unless /^READ:(.+)/;
        ($total_seqs, $avg_variants, $pv_size) = split ',', $1;
    }
    close $INFILE;


    # READ BIT COSTS FOR EACH OF THE ALGORITHMS
    my %run_cost;
    my %var_cost;
    for ( 'Elias-Gamma', 'Golomb', 'Huffman' ) {

        my $analysis_file = $analysis_dir;
        if ($chromosome) {
            $analysis_file .= 'hapmap_chr' . $chromosome;
        }
        else {
            $analysis_file .= $data_set;
        }
        $analysis_file .= '_' . $_ . '.txt';

        open $INFILE, '<', $analysis_file
            || die "Could not open file";
        while ( my $line = <$INFILE> ) {
            chomp $line;
            next unless $line =~ /Exp. Bit Cost:\s*(.+)/;
            if ( $run_cost{$_} ) {
                $var_cost{$_} = $1;
            }
            else {
                $run_cost{$_} = $1;
            }

        }
        close $INFILE;
    }


    # open the results file
    my $results_file = $analysis_dir;
    if ($chromosome) {
        $results_file .= 'hapmap_chr' . $chromosome;
    }
    else {
        $results_file .= $data_set;
    }
    $results_file .= '_compression_factors.txt';

    open my $OUTFILE, '>', $results_file
        || die "Could not open file";


    # output the general results info
    print $OUTFILE "Compression Factors - $data_set";
    if ($chromosome) {
        print $OUTFILE " - chromosome $chromosome";
    }
    print $OUTFILE "\n\n";
    print $OUTFILE "Number of seqs: $total_seqs\n";
    print $OUTFILE "Average # of variable positions per seq: $avg_variants\n\n";


    # output results with respect to full seqs
    print $OUTFILE "Compression factors with respect to full sequences\n";
    printf $OUTFILE "full seq. bit size: %.4f\n", $file_size;
    print $OUTFILE "\tassumes 15446 bases per coding-region sequence and 16568 for full\n";
    print $OUTFILE "\t* sequence identifiers ignored\n";
   for ( 'Elias-Gamma', 'Golomb', 'Huffman' ) {

        my $encoded_size = ( $run_cost{$_} + $var_cost{$_} ) * $avg_variants * $total_seqs;
        my $compression_factor = $file_size / $encoded_size;
        print $OUTFILE "$_:\n";
        printf $OUTFILE "\t%15s %.4f\n", 'Bit Size:', $encoded_size;
        printf $OUTFILE "\t%15s %.4f\n\n", 'Comp Factor:', $compression_factor;
    }
    print $OUTFILE "\n";


    # results with respect to pv representation
    print $OUTFILE "Compression factors with respect to pos-variant representation\n";
    printf $OUTFILE "pos-variant bit size: %.4f\n", $pv_size;
    print $OUTFILE "\t* sequence identifiers ignored\n";
    print $OUTFILE "\t* one new-line character per pos-variant pair\n";
    for ( 'Elias-Gamma', 'Golomb', 'Huffman' ) {

        my $encoded_size = ( $run_cost{$_} + $var_cost{$_} ) * $avg_variants * $total_seqs;
        my $compression_factor = $pv_size / $encoded_size;
        print $OUTFILE "$_:\n";
        printf $OUTFILE "\t%15s %.4f\n", 'Bit Size:', $encoded_size;
        printf $OUTFILE "\t%15s %.4f\n\n", 'Comp Factor:', $compression_factor;
    }

    close $OUTFILE;

}#FOLDEND


sub analyze_encoding {#FOLDBEG
    my ( $data_set, $chromosome, $algorithm ) = @_;


    #Create a filehandle for the results
    my $results_file = $analysis_dir . $data_set;
    $results_file .= '_chr'.$chromosome if $chromosome;
    $results_file .= '_' . $algorithm . '.txt';
    open my $RESULTS, '>', $results_file
        || die "Could not open file: $results_file";

    #Create a couple of stubs for defining the data files
    my $freq_stub = $frequency_dir . $data_set . '/';
    $freq_stub .= 'chr' . $chromosome . '_' if $chromosome;
    my $encoding_stub = $encoding_dir .  $data_set . '/';
    $encoding_stub .= 'chr' . $chromosome . '_' if $chromosome;


    #Analyze the runlengths
    print $RESULTS "$data_set - ";
    print $RESULTS "chromosome_$chromosome - " if $chromosome;
    print $RESULTS "$algorithm - ";
    print $RESULTS "runlengths\n";

    my $freq_file = $freq_stub . 'runlength.txt';
    die "Expected $freq_file but not found" unless -e $freq_file;
    my $encoding_file = $encoding_stub . 'runlength_' . $algorithm . '.txt';
    die "Expected $encoding_file but not found" unless -e $freq_file;

    open my $FREQS, '<', $freq_file 
        || die "Could not open file: $freq_file";
    open my $ENCODINGS, '<', $encoding_file 
        || die "Could not open file: $encoding_file";

    my $data = get_analysis_data( $FREQS, $ENCODINGS );
    compute_encoding_results( $data, $RESULTS );
    print $RESULTS "\n";

    close $FREQS;
    close $ENCODINGS;

    
    #Analyze the variants
    $freq_file = $freq_stub . 'variant.txt';
    $encoding_file = $encoding_stub . 'variant_' . $algorithm . '.txt';
    print $RESULTS "$data_set - ";
    print $RESULTS "chromsome$chromosome - " if $chromosome;
    print $RESULTS "$algorithm - ";
    print $RESULTS "variants\n";

    open $FREQS, '<', $freq_file 
        || die "Could not open file: $freq_file\n";
    open $ENCODINGS, '<', $encoding_file 
        || die "Could not open file: $encoding_file\n";

    $data = get_analysis_data( $FREQS, $ENCODINGS );
    compute_encoding_results( $data, $RESULTS );

    close $FREQS;
    close $ENCODINGS;

    close $RESULTS;
    return $results_file;
}#FOLDEND


sub analyze_encoding_hapmap_pv {#FOLDBEG

    #Analyzes the position-variant representation for hapmap data of each chromosome
    print "Computing position-variant stats on HapMap sequences\n";

    #Analyze all the chromsome data that has been extracted
    # Sizes are computed using bit values
    my $analysis_file;
    for ( (1 .. 22), 'X', 'Y' ) {
        my $consensus_file = $reference_dir . 'chr' . $_ . 
            '_hapmap_consensus.txt';

        if ( -e $consensus_file ) { 

            #Get the total number of donors, average size of the donor indentifier, 
            #average size of a variant position, and the average number of variants 
            #per donor.
            my ( $total_donors, $avg_variants, $avg_ident_length, $avg_pos_length, $avg_var_length ) = 
                get_seq_stats_hapmap($_);


            #Compute the space requirements.  8 bits are added to the ID line
            #for the '>' and a newline.  24 bits are added to each variant
            #line - 16 for the 2-character variant and 8 for a newline.
            my $ident_line = 8 + ($avg_ident_length * 8) + 8;
            my $variant_lines = $avg_variants * ( ( ($avg_pos_length + $avg_var_length) * 8) + 8 );

            # Ignore the identification line for computing the expected seq
            # size.  Add the $ident_line here if it should be included.
            my $exp_seq_size = $variant_lines;

            #Print out a report
            $analysis_file = $analysis_dir . 'hapmap_chr' . $_ . '_pv.txt';
            open my $ANALYSIS, '>', $analysis_file
                || die "Could not open file: $analysis_file";

            print $ANALYSIS "hapmap - chromosome_$_ - pos-variant\n";
            printf $ANALYSIS "\t%-40s%6u\n", 'Total Seqs:', $total_donors;
            printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # variants per seq:', $avg_variants;
            printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # characters per variant:', $avg_var_length;
            printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # characters per position:', $avg_pos_length;
            printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # characters per seq identifier:', $avg_ident_length;
            print $ANALYSIS "Expected size per sequence using position-variant notation:\n";
            print $ANALYSIS "** assumes one newline character following each position-variant pair\n";
            print $ANALYSIS "** sequence identifiers not included in the calculation\n";
            print $ANALYSIS '** per seq. bit size = $avg_variants * ( ( ($avg_pos_length + $avg_var_length) * 8) + 8 )', "\n";
            printf $ANALYSIS "\t%-15s%.4f\n", 'bits:', $exp_seq_size;
            printf $ANALYSIS "\t%-15s%.4f\n", 'bytes:', $exp_seq_size / 8;
            printf $ANALYSIS "\t%-15s%.4f\n", 'K-bytes:', $exp_seq_size / ( 8 * 1024 );
            print $ANALYSIS "Expected file size for all sequences:\n";
            print $ANALYSIS "** values per sequence multiplied by the number of sequences\n";
            printf $ANALYSIS "\t%-15s%.4f\n", 'bits:', $exp_seq_size * $total_donors;
            printf $ANALYSIS "\t%-15s%.4f\n", 'bytes:', ($exp_seq_size / 8) * $total_donors;
            printf $ANALYSIS "\t%-15s%.4f\n", 'K-bytes:', ($exp_seq_size / ( 8 * 1024 )) * $total_donors;
            printf $ANALYSIS "\t%-15s%.4f\n\n\n", 'M-bytes:', ($exp_seq_size / ( 8 * 1024 * 1024 )) * $total_donors;
            print $ANALYSIS '#' x 40, ' IGNORE VALUES BELOW ', '#' x 40,"\n";
            print $ANALYSIS "READ:", $total_donors, ',', $avg_variants, ',',
                $exp_seq_size * $total_donors;

            #Values written on the last line of this file are used by other
            #subroutines to compute the compression ratio.

            close $ANALYSIS;
        }
        return $analysis_file;

    }
}#FOLDEND


sub analyze_encoding_pv {#FOLDBEG
    my $data_set = shift;

    #Analyzes the position-variant representation of mtDNA data 
    print "Computing position-variant stats on $data_set sequences\n";

    #Get the total number of donors, average size of the donor indentifier, 
    #average size of a variant position, and the average number of variants 
    #per donor.
    my ( $total_donors, $avg_variants, $avg_ident_length, $avg_pos_length, $avg_var_length ) = 
        get_seq_stats($data_set);


    #Compute the space requirements.  8 bits are added to the ID line
    #for the '>' and a newline.  24 bits are added to each variant
    #line - 16 for the 2-character variant and 8 for a newline.
    my $ident_line = 8 + ($avg_ident_length * 8) + 8;
    my $variant_lines = $avg_variants * ( ( ($avg_pos_length + $avg_var_length) * 8) + 8 );

    # Ignore the identification line for computing the expected seq
    # size.  Add the $ident_line here if it should be included.
    my $exp_seq_size = $variant_lines;

    #Print out a report
    my $analysis_file = $analysis_dir . $data_set . '_pv.txt';
    open my $ANALYSIS, '>', $analysis_file
        || die "Could not open file: $analysis_file";

    print $ANALYSIS "$data_set - pos-variant\n";
    printf $ANALYSIS "\t%-40s%6u\n", 'Total Seqs:', $total_donors;
    printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # variants per seq:', $avg_variants;
    printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # characters per variant:', $avg_var_length;
    printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # characters per position:', $avg_pos_length;
    printf $ANALYSIS "\t%-40s%.4f\n", 'Avg. # characters per seq identifier:', $avg_ident_length;
    print $ANALYSIS "Expected size per sequence using position-variant notation:\n";
    print $ANALYSIS "** assumes one newline character following each position-variant pair\n";
    print $ANALYSIS "** sequence identifiers not included in the calculation\n";
    print $ANALYSIS '** per seq. bit size = $avg_variants * ( ( ($avg_pos_length + $avg_var_length) * 8) + 8 )', "\n";
    printf $ANALYSIS "\t%-15s%.4f\n", 'bits:', $exp_seq_size;
    printf $ANALYSIS "\t%-15s%.4f\n", 'bytes:', $exp_seq_size / 8;
    printf $ANALYSIS "\t%-15s%.4f\n", 'K-bytes:', $exp_seq_size / ( 8 * 1024 );
    print $ANALYSIS "Expected file size for all sequences:\n";
    print $ANALYSIS "** values per sequence multiplied by the number of sequences\n";
    printf $ANALYSIS "\t%-15s%.4f\n", 'bits:', $exp_seq_size * $total_donors;
    printf $ANALYSIS "\t%-15s%.4f\n", 'bytes:', ($exp_seq_size / 8) * $total_donors;
    printf $ANALYSIS "\t%-15s%.4f\n", 'K-bytes:', ($exp_seq_size / ( 8 * 1024 )) * $total_donors;
    printf $ANALYSIS "\t%-15s%.4f\n\n\n", 'M-bytes:', ($exp_seq_size / ( 8 * 1024 * 1024 )) * $total_donors;
    print $ANALYSIS '#' x 40, 'IGNORE VALUES BELOW', '#' x 40,"\n";
    print $ANALYSIS "READ:", $total_donors, ',', $avg_variants, ',', $exp_seq_size * $total_donors;

    #Values written on the last line of this file are used by other
    #subroutines to compute the compression ratio.

    close $ANALYSIS;
    return $analysis_file;

}#FOLDEND


sub compute_encoding_results {#FOLDBEG
    my ( $data, $RESULTS ) = @_;

    my $count;          #number of symbols
    my $exp_value;      #exp. value of the symbols when numbered 1 to $count
    my $exp_bit_cost;   #expected bit cost for a symbol in that encoding
    my $sum_nsq_freq;   #sum(n**2 * freq)
    my $var;            #variance
    my $std_dev;        #standard deviation

    my $i = 1;
    for ( @{$data} ) {
        my ( $p, $f, $c ) = split ',', $_;
        
        $count++;
        $exp_value += $i * $f;
        $exp_bit_cost += $f * length($c);
        $sum_nsq_freq += $i**2 * $f;

        $i++;
    }

    $var = $sum_nsq_freq - $exp_value**2;
    $std_dev = sqrt $var;


    printf $RESULTS "\t%-25s%6u\n", 'Symbols:', $count;
    printf $RESULTS "\t%-25s%.6f\n", 'Exp. n-Value:', $exp_value;
    printf $RESULTS "\t%-25s%.6f\n", 'Exp. Bit Cost:', $exp_bit_cost;
    printf $RESULTS "\t%-25s%.6f\n", 'Variance:', $var;
    printf $RESULTS "\t%-25s%.6f\n", 'Std. Dev:', $std_dev;

}#FOLDEND


sub compute_hapmap_frequencies {#FOLDBEG

    for ( (1 .. 22), 'X', 'Y' ) {
        my $consensus_file = $reference_dir . 'chr' . $_ . 
            '_hapmap_consensus.txt';

        if ( -e $consensus_file ) { 

            #Compute the runlength frequency
            compute_hapmap_runlength_frequencies($_);

            #Compute the variant frequency
            compute_hapmap_variant_frequencies($_);

        }

    }

}#FOLDEND


sub compute_hapmap_runlength_frequencies {#FOLDBEG
    my ( $chromosome ) = @_;

    print "Computing runlength frequencies for chromosome $chromosome\n";

    #Get a connection to the database
	my $dbh = get_dbh( $chromosome );

    #Prepare a statement for getting all the positions associated with a donor
    my $donor_variant_stm = $dbh->prepare( qq{
        SELECT position
        FROM donor, variant, donor_variant
        WHERE
            donor.id = donor_variant.donor_id AND
            variant.id = donor_variant.variant_id AND
            donor.id = ?
        ORDER BY position ASC
    });


    #Get a list of the donor IDs - This query is very simple because every
    #chromosome has its own sqlite database, so all of the donors in this
    #particular database are associated with this chromosome.
    my $donor_ids = $dbh->selectall_arrayref ( qq{
        SELECT id
        FROM donor
        });

    
    #Loop through the donors, convert positions to runlengths, and tally the
    #counts
    my %runlength_counts;
    my $total_count = 0;
    for ( @{$donor_ids} ) {
        print "\tTallying runlengths for donor ", $_->[0], "\n";

        $donor_variant_stm->execute($_->[0]);
        my $row_ref = $donor_variant_stm->fetchall_arrayref();

        my $current_position = 0;
        for ( @{$row_ref} ) {

            $runlength_counts{$_->[0] - $current_position}++;
            $total_count++;
            $current_position = $_->[0];
        }
    }


    #Write the counts and computed frequencies to files
    my $count_file = $count_dir . 'hapmap/chr' . $chromosome . '_runlength.txt';
    open my $COUNTS, '>', $count_file ||
        die "Could not open count_file: $count_file";

    my $frequency_file = $frequency_dir . 'hapmap/chr' . $chromosome . '_runlength.txt';
    open my $FREQUENCIES, '>', $frequency_file ||
        die "Could not open frequency_file: $frequency_file";

    for ( sort ( { $runlength_counts{$b} <=> $runlength_counts{$a} } keys %runlength_counts ) ) {
       printf $COUNTS "%-14u%12u\n", $_, $runlength_counts{$_};
       printf $FREQUENCIES "%-14u%.8f\n", $_, $runlength_counts{$_} / $total_count;    
    }

    close $COUNTS;
    close $FREQUENCIES;
}#FOLDEND


sub compute_hapmap_variant_frequencies {#FOLDBEG
    my ( $chromosome ) = @_;

    print "Computing variant frequencies for chromosome $chromosome\n";

    #Get a connection to the database
	my $dbh = get_dbh( $chromosome );


    #Tally the counts for each genotype
	my $row_ref = $dbh->selectall_arrayref ( qq{
	        SELECT variant, count(variant)
	        FROM donor, variant, donor_variant
	        WHERE
	            donor.id = donor_variant.donor_id AND
	            variant.id = donor_variant.variant_id
	        GROUP BY variant 
	        ORDER BY count(variant) DESC
	});
    my %genotype_counts;
    my $total_count = 0;
    map { $genotype_counts{$_->[0]} = $_->[1]; $total_count += $_->[1] } @{$row_ref};

    #Write the counts and computed frequencies to files
    my $count_file = $count_dir . 'hapmap/chr' . $chromosome . '_variant.txt';
    open my $COUNTS, '>', $count_file ||
        die "Could not open count_file: $count_file";

    my $frequency_file = $frequency_dir . 'hapmap/chr' . $chromosome . '_variant.txt';
    open my $FREQUENCIES, '>', $frequency_file ||
        die "Could not open variant_file: $frequency_file";

    for ( sort (  { $genotype_counts{$b} <=> $genotype_counts{$a} } keys %genotype_counts ) ) {
       printf $COUNTS "%-4s%12u\n", $_, $genotype_counts{$_};    
       printf $FREQUENCIES "%-4s%.8f\n", $_, $genotype_counts{$_} / $total_count;    
    }

    close $COUNTS;
    close $FREQUENCIES;
}#FOLDEND


sub compute_frequencies {#FOLDBEG
    my ($data_set) = @_;

    my $data_file = $data_dir . $data_set . '/variants.fasta';
    my $runlength_count_file = $count_dir . $data_set . '/runlength.txt';
    my $variant_count_file = $count_dir . $data_set . '/variant.txt';
    my $runlength_freq_file = $frequency_dir . $data_set . '/runlength.txt';
    my $variant_freq_file = $frequency_dir . $data_set . '/variant.txt';
    open my $DATA, '<', $data_file ||
        die "Could not open file";

    die "Expected file: $data_file but did not find it" unless -e $data_file;

    # Get the runlength and variant counts
    my %run_counts;
    my %var_counts;
    my $total_count = 0;
    my $current_genome;
    my $current_position;
    while ( <$DATA> ){

        # ignore blank lines
        next if /^\s*$/;
        
        chomp;

        if ( /^>(\w+)/ ) {
            $current_genome = $1;
            $current_position = 0;
            next;
        }

        my ( $r ) = /^\s*(\d+)/;
        $run_counts{ $r - $current_position }++;
        $current_position = $r;
        my $v = $';
        $var_counts{$v}++;
        $total_count++;
    }


    # Compute the runlength frequencies
    my %run_freqs;
    for ( keys %run_counts ) {
        $run_freqs{$_} = $run_counts{$_} / $total_count;
    }

    # Compute the variant frequencies
    my %var_freqs;
    for ( keys %var_counts ) {
        $var_freqs{$_} = $var_counts{$_} / $total_count;
    }
    
    

    # Write the values to files
    open my $RUNCOUNT, '>', $runlength_count_file ||
        die "Could not open file";
    for ( sort { $run_counts{$b} <=> $run_counts{$a} } keys %run_counts ) {

        printf $RUNCOUNT "%-10u %-6u\n", $_, $run_counts{$_};

    }
    close $RUNCOUNT;
    

    open my $VARCOUNT, '>', $variant_count_file ||
        die "Could not open file";
    for ( sort { $var_counts{$b} <=> $var_counts{$a} } keys %var_counts ) {

        printf $VARCOUNT "%-15s %-6u\n", $_, $var_counts{$_};

    }
    close $VARCOUNT;


    open my $RUNFREQ, '>', $runlength_freq_file ||
        die "Could not open file";
    for ( sort { $run_freqs{$b} <=> $run_freqs{$a} } keys %run_freqs ) {

        printf $RUNFREQ "%-10u %.8f\n", $_, $run_freqs{$_};

    }
    close $RUNFREQ;



    open my $VARFREQ, '>', $variant_freq_file ||
        die "Could not open file";
    for ( sort { $var_freqs{$b} <=> $var_freqs{$a} } keys %var_freqs ) {

        printf $VARFREQ "%-10s %.8f\n", $_, $var_freqs{$_};

    }
    close $VARFREQ;

}#FOLDEND


sub download_hapmap_files {#FOLDBEG
	#Configure the user agent pretending to be mozilla
	my $ua = LWP::UserAgent->new();
	$ua->agent('Mozilla/5.0 (Macintosh; U; PPC Mac OS X Mach-O; en-US; rv:1.8.0.9) Gecko/20061206 Firefox/1.5.0.9');
	
	#Generate the list of files to be download
	my $base_url = 'http://ftp.hapmap.org/genotypes/latest/forward/non-redundant/';
	my $res = $ua->get($base_url);
	my @res_lines = split "\n", $res->content();
	my @genotype_files = map { /(genotypes_.+?\.gz)/ } @res_lines;
	
	#Download the files
	foreach my $file_name (@genotype_files) {
	    my $file_url = $base_url . $file_name;
	
	    #retrieve the response
	    my $res = $ua->get($file_url,':content_file'=>'data/hapmap/'.$file_name);
        print "Downloaded $file_name\n";
	}
}#FOLDEND


sub encode_elias_gamma {#FOLDBEG

    my $n = shift;

    #Do some error checking
    if ( $n < 0 ) {
        die "parameter to encode_elias_gamma cannot be less than 0"
    }

    #Increment $n so that we can begin our encoding at zero
    $n++;

    #Generate the prefix
    my $prefix;
    my $k;
    if ($n == 0) {
        $k = 0;
    }
    else {
        $k = int(log($n) / log(2));
    }
    $prefix =  'O' x $k;

    #Generate the suffix
    my $suffix;
    my $i = 0;
    do {
        my $temp = $n % 2;

        if ( $temp > 0 ) {
            if ( $suffix ) {
                $suffix = 'l' . $suffix;
            }
            else {
                $suffix = 'l';
            }
        }
        else {
            if ( $suffix ) {
                $suffix = 'O' . $suffix;
            }
            else {
                $suffix = 'O';
            }
        }

        $n = int($n/2);
    } until ( $n == 0 );

    my $code;
    if ($suffix) {
        $code = $prefix . $suffix;
    }
    else {
        $code = $prefix;
    }

    return $code;
}#FOLDEND


sub encode_golomb {#FOLDBEG
    my ($n, $m) = @_;

    #Do some error checking
    if ($m < 1) {
        die "The second parameter to encode_golomb cannot be less than one";
    }

    my $q = int($n/$m);

    my $scale = encode_unary($q);

    #Get the mantissa and its encoding
    my $r = $n - ($q * $m);

    my $mantissa = encode_truncated_binary($m, $r);

    my $rv = $mantissa ? $scale . $mantissa : $scale;
    return($rv);

}#FOLDEND


sub encode_truncated_binary {#FOLDBEG
    my ($n, $r) = @_;
    # $n is the total alphabet size 
    # $r is the symbol in the alphabet to encode

    die "Second parameter must be an integer >= 0."
        if ( $r < 0 );

    die "Second parameter cannot be larger than the first for truncated-binary."
        if ( $r > $n );

    #Increment $r so that we can generate symbols beginning at zero
    $r++;

    #Set the k-value 
    my $k = log($n) / log(2);
    if ( ($k - int($k)) > 0  ) {
        $k = int($k);
        $k++;
    }
    else {
        $k = int($k);
    }
 
    #Determine the number of symbols that can be encoded with k-1 bits
    #(truncated encodings)
    my $t = $n - ( 2 * ( $n - 2**($k-1) ) );
    my $s = 2**$k -$n;
   
    #Set the length for the code and the binary value that will be encoded
    my $code_length;
    my $b;
    if ( $r <= $t ) {
        $code_length = $k - 1;
        $b = $r - 1;
    }
    else {
        $code_length = $k;
        $b = $r + $s - 1;
    }

    #Create the code
    my $code;
    for (my $i = ($code_length - 1); $i >= 0; $i--) {
        if ( (2**$i) <= $b ) {
            $code .= 'l';
            $b = $b - (2**$i);
        }
        else {
            $code .= 'O';
        }
    }

    return $code;
}#FOLDEND


sub encode_unary {#FOLDBEG
    my ($n) = @_;

    die "n can't be less than zero for unary encoding"
        if $n < 0;

    return ('l' x $n) . 'O';
}#FOLDEND


sub extract_hapmap_genotype_data {#FOLDBEG
    #Extracts the individual genotype values listed in the hapmap files and
    #stores them in a database.  Also generates a text file of the consensus
    #value at each variable position using position-value nomenclature.

    print "Extracting haplotype data\n";

    #create lists of the data file names by chromosome
    opendir my $DATADIR, $data_dir.'hapmap' ||
        die "Could not open hapmap data file.";

    my %file_lists;
    my $chromosome;
    my $population;
    while (my $file_name = readdir $DATADIR){
        next if ($file_name eq '.' or $file_name eq '..');
        ($chromosome, $population) = $file_name =~ /genotypes_chr([^_]{1,2})_([A-Z]{3})_.+?\.gz/;   
        push @{ $file_lists{$chromosome} }, $data_dir.'hapmap/'.$file_name;

        #get rid of the previous consensus files
        my $consensus_file = $reference_dir . 'chr'
            . $chromosome . '_hapmap_consensus.txt';
        unlink $consensus_file if -e $consensus_file;
    }
    closedir $DATADIR;


    #process the data files
    foreach $chromosome (keys %file_lists) {

        my %data_grid;
        my @donor_list;
        #make a position-donor grid of all the data for the chromosome
        print "Processing chromosome $chromosome\n";
        get_chromosome_data( $chromosome, $file_lists{$chromosome}, \%data_grid, \@donor_list );

        #make adjustments for the most frequent haplotype at each position
        make_consensus_adjustments( $chromosome, \%data_grid );

        #store the adjusted data
        store_hapmap_data( $chromosome, $population, \%data_grid, \@donor_list );

    }

    return;
}#FOLDEND


sub get_analysis_data {#FOLDBEG
    my ( $FREQS, $ENCODINGS ) = @_;

    my @data;
    while ( my $freq_line = <$FREQS> ) {
        my $enc_line = <$ENCODINGS>;

        my ( $p, $f ) = split /\s+/, $freq_line;
        my ( $q, $e ) = split /\s+/, $enc_line;

        push @data, join ',', ($p, $f, $e);

    }
    return \@data;
}#FOLDEND


sub get_chromosome_data {#FOLDBEG
    my ( $chromosome, $file_list, $data_grid, $donor_list ) = @_;

    open my $OUTFILE, '>', $analysis_dir.'/hapmap_chr'.$chromosome.'_files.txt'
        || die "Could not open outfile";

    print $OUTFILE "HapMap File Sizes - Chromosome $chromosome\n";

    my $zipped_total = 0;
    my $unzipped_total = 0;


    for ( @{ $file_list } ) {

        $zipped_total += -s $_;
        #unzip the file
        system $GUNZIP, $_; 
        my $unzipped_name = $_;
        chop $unzipped_name;
        chop $unzipped_name;
        chop $unzipped_name;
        $unzipped_total += -s $unzipped_name;

        #extract the data from the file
        get_hapmap_file_data( $unzipped_name, $data_grid, $donor_list );

        #rezip the file
        system $GZIP, $unzipped_name; 
    }

    print $OUTFILE @{ $file_list }, " Files\n\n";
    print $OUTFILE "Zipped Size:\n";
    printf $OUTFILE "\t%10s %.2f\n", 'bits:', ($zipped_total * 8);
    printf $OUTFILE "\t%10s %.2f\n", 'bytes:', $zipped_total;
    printf $OUTFILE "\t%10s %.2f\n", 'K-bytes:', $zipped_total / 1024;
    printf $OUTFILE "\t%10s %.2f\n\n", 'M-bytes:', $zipped_total / ( 1024 * 1024 );

    print $OUTFILE "Unzipped Size:\n";
    printf $OUTFILE "\t%10s %.2f\n", 'bits:', ($unzipped_total * 8);
    printf $OUTFILE "\t%10s %.2f\n", 'bytes:', $unzipped_total;
    printf $OUTFILE "\t%10s %.2f\n", 'K-bytes:', $unzipped_total / 1024;
    printf $OUTFILE "\t%10s %.2f", 'M-bytes:', $unzipped_total / ( 1024 * 1024 );

    # Add some numbers needed for compression factors.  This is not the ideal
    # location to put them, but the functions are already reading these
    # values.  We want the size for a complete chromosme represent.
    my %size_lookup = (
        1 => 247000000,
    );

    print $OUTFILE '#' x 40, ' IGNORE VALUES BELOW ', '#' x 40,"\n";
    print $OUTFILE "READ:", $size_lookup{$chromosome} * 8;
    close $OUTFILE;

}#FOLDEND


sub get_mtdna_seq_ids {#FOLDBEG

    my $id_file = $data_dir . '/mtdna_rcrs/genome_ids.txt';
    open my $INFILE, '<', $id_file
    || die "Could not open infile";

    my @seq_ids;
    while ( <$INFILE> ) {
        chomp;
        push @seq_ids, $_;
    }

    return \@seq_ids;
}#FOLDEND


sub get_variant_lists {#FOLDBEG
    my ( $data_set ) = @_;

    my $data_file = $data_dir . $data_set . '/variants.fasta';
    die "Datafile: $data_file is not present" unless -e $data_file;
    open my $DATAFILE, '<', $data_file
        || die "Could not open data file";

    my %var_lists;
    my $current_genome;
    my @current_list;
    while ( <$DATAFILE> ) {
        chomp;
        if ( /^>(.+)/ ) {
            if ($current_genome) {
                my @new_list = @current_list;
                $var_lists{$current_genome} = \@new_list;
                @current_list = ();
            }
            $current_genome = $1;
        }
        else {
            push @current_list, $_;
        }
    }
    $var_lists{$current_genome} = \@current_list;
                                   
   return \%var_lists;

}#FOLDEND


sub decompress {#FOLDBEG
    my ($file_name) = @_;


    # Get the donor identifier, algorithm, and dataset info
    # from the file name;
    my ($short_name) = $file_name =~ m{([^/]+)\.bit$};
    my ($identifier, $dataset, $algorithm) = split '__', $short_name;


    # Set the encoding file names
    my $runlength_file;
    my $variant_file;
    if ( $dataset =~ /^hapmap_chr(\d+)/ ) {
       $runlength_file = $encoding_dir . 'hapmap/chr'. $1 . '_' . 'runlength' 
            . '_' . $algorithm . '.txt'; 

       $variant_file = $encoding_dir . 'hapmap/chr'. $1 . '_' . 'variant' 
            . '_' . $algorithm . '.txt'; 
    }
    else {
       $runlength_file = $encoding_dir . $dataset . '/' . 'runlength' 
            . '_' . $algorithm . '.txt'; 

       $variant_file = $encoding_dir . $dataset . '/' . 'variant' 
            . '_' . $algorithm . '.txt'; 
    }

    #Retrieve the encodings
    my %runlength_encoding;
    open my $INFILE, '<', $runlength_file ||
        die "Could not open encoding file";
    while (my $line = <$INFILE>) {
        chomp $line;
        my ($r, $e) = split /\s+/, $line;
        $runlength_encoding{$r} = $e;
    }
    close $INFILE;
            
    my %variant_encoding;
    open $INFILE, '<', $variant_file ||
        die "Could not open encoding file";
    while (my $line = <$INFILE>) {
        chomp $line;
        my ($v, $e) = split /\s+/, $line;
        $variant_encoding{$v} = $e;
    }
    close $INFILE;


    #Slurp in the message - This file should have a one line bit string.
    open my $CODEFILE, '<', $file_name
        || die 'Could not open data file';

    my $bit_string = <$CODEFILE>;
    close $CODEFILE;


    #Decode the message
    my $current_position = 0;
    while ($bit_string) {
        no warnings;
        
        #Decode the runlength
        my $found;
        foreach my $k ( keys %runlength_encoding ) {
            $found = 0;
            if ($bit_string =~ /^$runlength_encoding{$k}/){

                my $p = $k + $current_position; 
                print "$p";

                $bit_string = substr( $bit_string, (length $runlength_encoding{$k}) ); 
                $current_position = $p;
                $found = 1;

                last;
            }
        }

        unless ( $found ) {
            die "Unknown bit sequence for runlength: $bit_string";
        }


        #Decode the variant
        foreach my $k ( keys %variant_encoding ) {
            $found = 0;
            if ($bit_string =~ /^$variant_encoding{$k}/){

                print "$k\n";

                $bit_string = substr( $bit_string, (length $variant_encoding{$k}) ); 
                $found = 1;

                last;
            }
        }

        unless ( $found ) {
            die "Unknown bit sequence for variant: $bit_string";
        }
    }
}#FOLDEND


sub get_seq_stats {#FOLDBEG

    my ( $data_set ) = @_;

    # Get the lists of variants
    my $variant_lists = get_variant_lists($data_set);

    my $total_donors = 0;
    my $total_variants = 0;
    my $ident_sum = 0;
    my $pos_sum = 0;
    my $var_sum = 0;
    for ( keys %{$variant_lists} ) {

        $total_donors++;
        $ident_sum += length($_);

        my $list = $variant_lists->{$_};
        for ( @{$list} ) {
            $total_variants++;
            my ( $p, $v ) = $_ =~ /^(\d+)(.+)/;
            $pos_sum += length($p);
            $var_sum += length($v);

        }
    }

    # Compute the averages (these values are in number of characters)
    my $avg_ident_length = $ident_sum / $total_donors;
    my $avg_pos_length = $pos_sum / $total_variants;
    my $avg_var_length = $var_sum / $total_variants;
    my $avg_variants = $total_variants / $total_donors;
   
    return ( $total_donors, $avg_variants, $avg_ident_length, $avg_pos_length, $avg_var_length ); 

}#FOLDEND


sub get_seq_stats_hapmap {#FOLDBEG
    #Generates basic stats for all the donors associated with a particular
    #chromosome.

    my ( $chromosome ) = @_;

    #Get a database connection
    my $dbh = get_dbh($chromosome);

    print "Computing donor count and avg. identifier length\n";

    # Count the number of donors
    my ($total_donors) = $dbh->selectrow_array( qq{
        SELECT count(*) FROM donor} );

    # Get the average length of an identifier
    my ($avg_ident_length) = $dbh->selectrow_array( qq{
        SELECT sum(length(code))/count(*) 
        FROM donor });
   

    # Get the total number of variants
    print "Getting count of variants\n";
    my ($total_variants) = $dbh->selectrow_array ( q{
        SELECT COUNT(*) FROM donor_variant });

    # Get the sum of characters used for all position values
    print "Getting the total number of characters used for positions\n";
    my ($pos_sum) = $dbh->selectrow_array ( q{
        SELECT SUM(LENGTH(position)) 
        FROM donor, variant, donor_variant 
        WHERE 
            donor.id = donor_variant.donor_id AND 
            variant.id = donor_variant.variant_id;
    });
    #my $pos_sum = 829984429;


    # Use a fixed value for the average variant length since all HapMap
    # haplotypes are two characters.  The query below will compute this value
    # if desired.
    my $avg_var_length = 2;
    # Get the sum of characters used for all variant values
    #print "Getting the total number of characters used for variants\n";
    #my ($var_sum) = $dbh->selectrow_array ( q{
    #    SELECT SUM(LENGTH(variant)) 
    #    FROM donor, variant, donor_variant 
    #    WHERE 
    #        donor.id = donor_variant.donor_id AND 
    #        variant.id = donor_variant.variant_id;
    #});
    #my $avg_var_length = $var_sum / $total_variants;


    my $avg_variants = ($total_variants / $total_donors);
    my $avg_pos_length = $pos_sum / $total_variants;


    return ( $total_donors, $avg_variants, $avg_ident_length, $avg_pos_length, $avg_var_length ); 
}#FOLDEND


sub get_key_data {#FOLDBEG
    my ( $data_set, $run_or_var, $count_or_freq) = @_;

    my %key_data;
    if ( $count_or_freq eq 'counts' ) {
	    my $count_file;
	    if ( $data_set eq 'mtdna_rcrs' ) {
	    
            compute_frequencies($data_set);
	        $count_file = $count_dir . 'mtdna_rcrs/' . $run_or_var . '.txt';

	    }
	    elsif ( $data_set eq 'mtdna_consensus' ) {
	
            compute_frequencies($data_set);
	        $count_file = $count_dir . 'mtdna_consensus/' . $run_or_var . '.txt';

	    }
        elsif ( $data_set eq 'your_data' ) {
	
            compute_frequencies($data_set);
	        $count_file = $count_dir . 'your_data/' . $run_or_var . '.txt';

	    }

	    else {
	
	        $count_file = $count_dir . 'hapmap/chr' .  $data_set . '_' . $run_or_var . '.txt';
	    }

        die "Expected file $count_file but did not find it" unless -e $count_file;

        # Read the counts
	    open my $COUNTS, '<', $count_file ||
	        die "Can't open count file: $count_file";
	    while ( <$COUNTS> ){
	        chomp;
	        my ( $k, $c ) = split /\s+/;
	        $key_data{$k} = $c;
	    }
    }

    if ( $count_or_freq eq 'freqs' ) {

	    my $freq_file;
	    if ( $data_set eq 'mtdna_rcrs' ) {
	    
	        $freq_file = $frequency_dir . 'mtdna_rcrs/' . $run_or_var . '.txt';
	    }
	    elsif ( $data_set eq 'mtdna_consensus' ) {
	
	        $freq_file = $frequency_dir . 'mtdna_consensus/' . $run_or_var . '.txt';
	    }
        elsif ( $data_set eq 'your_data' ) {
	
	        $freq_file = $frequency_dir . 'your_data/' . $run_or_var . '.txt';

	    }

	    else {
	
	        $freq_file = $frequency_dir . 'hapmap/chr' .  $data_set . '_' . $run_or_var . '.txt';
	    }
	
        die "Expected file $freq_file but did not find it" unless -e $freq_file;

        # Read the frequencies
	    open my $FREQS, '<', $freq_file ||
	        die "Can't open freq file: $freq_file";
	    while ( <$FREQS> ){
	        chomp;
	        my ( $k, $f ) = split /\s+/;
	        $key_data{$k} = $f;
	    }
    }

    return \%key_data;

}#FOLDEND


sub get_dbh {#FOLDBEG
    my ( $chromosome ) = @_;

    my $db_file = $sqlite_dir . 'chr' . $chromosome . '.db';
    my $dbh = DBI->connect("dbi:SQLite:dbname=$db_file","","");

    #Set a few optimizations
	$dbh->do( qq{ PRAGMA synchronous=OFF } );
	$dbh->do( qq{ PRAGMA cache_size=1000000 } );
	$dbh->do( qq{ PRAGMA journal_mode=OFF } );
	$dbh->do( qq{ PRAGMA count_changes=OFF } );

    return $dbh;
}#FOLDEND


sub get_dbh_mito {#FOLDBEG
    # Returns a database handle to the mito database

    # Preferentially use connection parameters used in the method call, 
    #otherwise use the values defined in the config file.	
    my $host = 'db1.bio.uci.edu';
    my $port = 5432;
    my $database = 'mito';
    my $user = 'mitouser';
    my $password = undef;

    my $connect_string = "dbname=$database;host=$host;port=$port";        
    my $dsn = "dbi:Pg:$connect_string";
    my %dbi_params = (
        AutoCommit=>0,
        PrintError=>0,
        RaiseError=>1,
        ShowErrorStatement=>1
    );

    my $dbh = DBI->connect($dsn, $user, $password, \%dbi_params)
        || warn "No database connection";

    return $dbh
}#FOLDEND


sub get_hapmap_file_data {#FOLDBEG
    my ( $file_name, $data_grid, $donor_list ) = @_;   

    print "Processing file $file_name\n";

    #Open the hapmap data file
    open my $DATAFILE, '<', $file_name
        || die "Could not open file: $file_name";

    #First line of the file contains metadata, including the donor
    #identifiers.
    my $line = <$DATAFILE>;
    my @cols = split /\s/, $line;
    my @old_donors = @{$donor_list};
    my @new_donors = @cols[11 .. $#cols];
    @{$donor_list} = ( @old_donors, @new_donors );


    #Process all the lines after the first line of the file
    my @sorted_positions = sort { $a <=> $b } keys %{ $data_grid };
    while ( $line = <$DATAFILE> ) {

        @cols = split /\s+/, $line; 
        my $p = $cols[3];

        while ( $sorted_positions[0] and $sorted_positions[0] < $p ) {
            my $q = shift @sorted_positions;
            $data_grid->{$q} .= ',' . join( ',', split ( '', '0' x scalar @new_donors ) );
        }

        if ( $sorted_positions[0] and $sorted_positions[0] == $p ) {
            #Position seen in a previous file so add the new values
            my $q = shift @sorted_positions;
            $data_grid->{$q} .= ',' . join (',', @cols[11 .. $#cols]);
            next;
        }

        #A new position
        if ( scalar @old_donors ) {
            $data_grid->{$p} = join( ',', split ( '', '0' x scalar @old_donors ) );
            $data_grid->{$p} .= ',' . join (',', @cols[11 .. $#cols]);
        }
        else {
            $data_grid->{$p} =  join (',', @cols[11 .. $#cols]);
        }
    }

    #Finish any remaining positions 
    for ( @sorted_positions ) {
        $data_grid->{$_} .= ',' . join( ',', split ( '', '0' x scalar @new_donors ) );
    }

    close $DATAFILE;

    return;
}#FOLDEND


sub get_mtDNA_variants {#FOLDBEG
    my ($reference) = @_;

    # Get a list of the seq IDs used
    my $seq_ids = get_mtdna_seq_ids();

    # Get a connection to the mito database
    my $dbh = get_dbh_mito();
    # Start with a master list of all the variants for the specified ids
    my $all_var_list;
    if ($reference eq 'rCRS'){
        $all_var_list = get_mtdna_variants_rcrs(ids=>$seq_ids, unknown=>0, 
            permanent=>1, temporary=>0);        
    }
    elsif ($reference eq 'consensus'){
        $all_var_list = get_mtdna_variants_consensus(ids=>$seq_ids, unknown=>0, 
            permanent=>1, temporary=>0);        
    }
    else{
        die "Unknown reference sequence";
    }


    # Get a count of full and coding-region seqs
    my %seq_check;
    map { $seq_check{$_}++ } @{$seq_ids};
    my $full_seq_count = 0;
    my $coding_seq_count = 0;

    for ( keys %seq_check ) {
        my $seq_attributes = $dbh->selectrow_hashref(qq{
            SELECT *
            FROM mitomaster.genome
            WHERE id = ?}, {}, $_);
        # The current seq is either full or coding-region only.
        if ($seq_attributes->{coverage} eq 'coding'){
            $coding_seq_count++;
        }
        else {
            $full_seq_count++;
        }
    }


    # Transform the notation adjust for seqs with no variation
    my @variant_lists;
    my $current_genome = $all_var_list->[0][0];
    my @current_list;
    for (@{$all_var_list}){
        if ($_->[0] != $current_genome){

            # Check for seqs that did not have any variants, but are on the
            # list of desired seqs.
            if ($_->[0] > ($current_genome + 1) and $seq_check{$current_genome + 1}){
                my $s = $current_genome + 1;
                push @variant_lists, [ [$s] ];
            }
            my @new_list = @current_list;
            push @variant_lists, \@new_list;
            @current_list = ();
            $current_genome = $_->[0];

        }
        push @current_list, $_;
    } 

    push (@variant_lists, \@current_list) if @current_list > 0;


    # Write the predicted full sequence sizes to a file
    my $data_file = $analysis_dir . 'mtdna_file_sizes.txt';
    open my $OUTFILE, '>', $data_file
        || die "Could not open data file.";

    my $file_size = ($coding_seq_count * 15446 ) + ($full_seq_count * 16568);
    print $OUTFILE "mtDNA File Sizes\n";
    print $OUTFILE "\t**assumes 15446 bases per coding-region sequence and 16568 for full\n";
    print $OUTFILE "\t**seq identifiers not included in calculation\n";
    print $OUTFILE "Full Seq. Count: $full_seq_count     Coding Seq. Count: $coding_seq_count\n";
    printf $OUTFILE "\t%10s %.2f\n", 'bits:', ($file_size * 8);
    printf $OUTFILE "\t%10s %.2f\n", 'bytes:', $file_size;
    printf $OUTFILE "\t%10s %.2f\n", 'K-bytes:', $file_size / 1024;
    printf $OUTFILE "\t%10s %.2f", 'M-bytes:', $file_size / ( 1024 * 1024 );
    print $OUTFILE '#' x 40, ' IGNORE VALUES BELOW ', '#' x 40,"\n";
    print $OUTFILE "READ:", $file_size * 8;
    close $OUTFILE;


    # Write the variants to a file.
    $data_file = $data_dir . 'mtdna_' . lc($reference) . '/variants.fasta';
    open my $DATAFILE, '>', $data_file
        || die "Could not open data file.";

    $current_genome = undef;
    foreach my $list ( @variant_lists ){

        for ( @{$list} ){
            if ( !$current_genome or $current_genome != $_->[0] ){
                $current_genome = $_->[0];
                print $DATAFILE ">$current_genome\n";
            }

            next unless $_->[3];

            if ($_->[3] eq 'insertion'){
                next unless $_->[2] =~ /^[ACGT]+$/;
                print $DATAFILE int($_->[1]).'Ins'.$_->[2]."\n";              
            }
            elsif ($_->[3] eq 'deletion'){
                next unless $_->[2] =~ /^-+$/;
                print $DATAFILE $_->[1].'Del'.length($_->[2]).'bp'."\n";          
            }
            else{
                next unless $_->[2] =~ /^[ACGT]$/;
                print $DATAFILE $_->[1].$_->[2]."\n";        
            }

        }
    }
    close $DATAFILE;

    return \@variant_lists;
}#FOLDEND


sub get_mtdna_variants_rcrs {#FOLDBEG
    my %variant_flags = @_;

    #the ids argument will, if defined, cause only variants 
    #from the genome_ids found in the array to be retrieved
    $variant_flags{ids} = undef unless $variant_flags{ids};
    
    #define some flags that allows us to select which variant types are returned
    $variant_flags{insertion} = 1 unless exists $variant_flags{insertion};
    $variant_flags{deletion} = 1 unless exists $variant_flags{deletion};
    $variant_flags{unknown} = 1 unless exists $variant_flags{unknown};
    $variant_flags{transition} = 1 unless exists $variant_flags{transition};
    $variant_flags{transversion} = 1 unless exists $variant_flags{transversion};
    
    #the pemanent and temporary and flags function independently - the appropriate
    #flag from the group above AND the applicable flag below must
    #be set for a variant to be included
    $variant_flags{permanent} = 1 unless exists $variant_flags{permanent};
    $variant_flags{temporary} = 0 unless $variant_flags{temporary};

    my $dbh = get_dbh_mito();
    
    my $id_string = '('. join (',', @{$variant_flags{ids}}) . ')' 
        if $variant_flags{ids};     

    #create a query for retrieving snps
    my $snp_query = qq{ 
            SELECT genome_id, genome_position, residue, type, haplogroup_major, 
                haplogroup_minor, gi, category, status       
                FROM mitomaster.dna_variants 
            WHERE 
                genome_position is not null AND
                (type = 'transition' OR
                type = 'transversion' OR
                type = 'unknown') };
                    
    if ($id_string){
        $snp_query .= "AND genome_id IN $id_string";
    }

    #get a snp list of all the variants requested - indels are not added here
    my $row_ref = $dbh->selectall_arrayref( $snp_query );

    #build the variant list up according to the type of variants
    #requested
    my @variant_list;
    for (@{$row_ref}){
        no warnings;
        if ($variant_flags{$_->[3]} and $variant_flags{$_->[8]}){
            push @variant_list, $_;            
        }
    }

    #Add any indels.
    if ($variant_flags{insertion} || $variant_flags{deletion}){
        #create a query for retrieving indels
        my $indel_query = qq{
            SELECT genome_id, genome_position, variant, type, haplogroup_major, 
                haplogroup_minor, gi, category, status       
            FROM mitomaster.complex_variants};

        if ($id_string){
            $indel_query .= " WHERE genome_id IN $id_string";
        }
                    
        $row_ref = $dbh->selectall_arrayref( $indel_query );
         
        #add the results that meet our criteria
        for (@{$row_ref}){
            no warnings;
            if ($_->[3] eq 'deletion'){
                push (@variant_list, $_) if $variant_flags{$_->[8]};
            }
            
            elsif ($_->[3] eq 'insertion'){
                if ($variant_flags{$_->[8]}){
                    push (@variant_list, $_) if $_->[2] =~ /^[ACGT]+$/;   
                }
            }
        }              
    }
    
    #sort the list by genome ID and genome position
    @variant_list = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} 
        @variant_list;
       
    return \@variant_list;
}#FOLDEND


sub get_mtdna_variants_consensus {#FOLDBEG
    my %flags = @_;

    # Get a database connection
    my $dbh = get_dbh_mito();


    # Get the consensus sequence
    my @consensus_seq;
    push @consensus_seq, undef;
    my $array_ref = $dbh->selectall_arrayref(q{
       SELECT * FROM mitomaster.sequence_consensus() 
    });
    map { push @consensus_seq, $_->[0] } @{$array_ref};


    # Create a list of variants that are caused solely by the consensus perspective
    #print "\tgenerating list of consensus perspective variants:\n";
    my %consensus_variants;
    for (1..$#consensus_seq){
        if ($REF_SEQ_UNC[$_] ne $consensus_seq[$_]){
            $consensus_variants{$_} = [ $REF_SEQ_UNC[$_], 
                get_var_type($consensus_seq[$_], $REF_SEQ_UNC[$_]) ];
                #print "\t\t$_ $consensus_variants{$_}->[0]  $consensus_variants{$_}->[1]\n"; 
        }
    }     

    
    # Get a list of the sequence IDs and make a checklist
    my $seq_list = $flags{ids};
    my %seq_check;
    map { $seq_check{$_}++ } @{$seq_list};
    #print "\tprocessing the variation in ", scalar @{$seq_list}, " genomes\n";


    # Get the rCRS variants
    my $rcrs_variants = get_mtdna_variants_rcrs(%flags);
    #print "\tadjusting ", scalar @{$rcrs_variants}, " sequence variants\n";


    my $current_genome;
    my %consensus_variants_copy = %consensus_variants;
    my @adjusted_variants;

    for (my $i = 0; $i < @{$rcrs_variants}; $i++){
        my $rv = $rcrs_variants->[$i];

        $current_genome = $rv->[0] unless $current_genome;
        if ($current_genome != $rv->[0]){

            #we've reached a new genome - add the derived variants that are left

            foreach my $p ( keys %consensus_variants_copy ){
                if ( ($rv->[4] eq 'coding') 
                    && ( $p < 577 or $p > 16022 ) ){
                    next;
                }
                
                push @adjusted_variants,
                    [ $current_genome, $p, $consensus_variants_copy{$p}->[0], 
                        $consensus_variants_copy{$p}->[1], $consensus_variants_copy{$p}->[4] ];
            }
            
            #adust the current_genome and reset the consensus variants copy
            delete $seq_check{$current_genome};
            $current_genome = $rv->[0];
            %consensus_variants_copy = %consensus_variants;
        }

        if ($rv->[3] eq 'deletion' or $rv->[3] eq 'insertion'){
            #indels are included unchanged
            push @adjusted_variants, [ $current_genome, $rv->[1], $rv->[2], $rv->[3], $rv->[4] ];
            next;
        }

        if ($REF_SEQ_UNC[$rv->[1]] eq $consensus_seq[$rv->[1]]){
            #the consensus matches the reference at this position, so the variant
            #is the same
            push @adjusted_variants, [ $current_genome, $rv->[1], $rv->[2], $rv->[3], $rv->[4] ];
            next;            
        }

        if ($consensus_seq[$rv->[1]] eq $rv->[2]){
            #the variant is the same as the consensus change, so remove it as a derived variant
            delete $consensus_variants_copy{$rv->[1]};
            next;
        }  
        
        if ( ($REF_SEQ_UNC[$rv->[1]] eq $consensus_seq[$rv->[1]]) &&
             ($consensus_seq[$rv->[1]] ne $rv->[2]) &&
             ($REF_SEQ_UNC[$rv->[1]] ne $rv->[2]) ) {
            #the variant is at a position where the consensus differs from the reference,
            #but the variant residue also differs from the consensus

            push @adjusted_variants,
                [ $current_genome, $rv->[1], $rv->[2], 
                get_var_type($rv->[2], $consensus_variants{$rv->[1]}->[0]), $rv->[4] ];
            
            delete $consensus_variants_copy{$rv->[1]};
            next;
        }   
    }


    # Adjust sequences that had no variants in the variation list
    for (keys %seq_check){
        foreach my $p (keys %consensus_variants){
           push @adjusted_variants,
            [ $_, $p, $consensus_variants{$p}->[0], $consensus_variants{$p}->[1], '' ];
            
        }
    }
            

    # Sort the list
    @adjusted_variants = sort {$a->[0] <=> $b->[0] || $a->[1] <=> $b->[1]} 
        @adjusted_variants;
       
    return \@adjusted_variants;

}#FOLDEND


sub get_sequence_variants {#FOLDBEG

    my ($data_file, $seq_id) = @_;

    open my $INFILE, '<', $data_file
        || die "Could not open file";

    my $found = 0;
    my @var_ref;
    while ( <$INFILE> ) {

        if ( /^>/ and $found ) {
            # We are done
            last;
        }


        if ( /^>(.+)/ and !$found ) {
            # Check to see if this is our sequence
            if ($seq_id eq $1) {
                $found = 1;
                next;
            } 
        }


        if ( $found ) {
            my @vals = /(\d+)(.+)/;
            push @var_ref, \@vals;
        }
        else {
            next;
        }

    }

    close $INFILE;
    return \@var_ref;
}#FOLDEND


sub get_var_type {#FOLDBEG
    
    unless ($_[0] and $_[1]){
        warn "Determining a variant type requires two values";
        return;
    }
    
    if ($_[0] eq $_[1]){
        warn "No var type for matching residues: $_[0] $_[1]";
        return;
    }
    
    my %v_type = (
        AC=>'transversion',
        AG=>'transition',
        AT=>'transversion',
        CA=>'transversion',
        CG=>'transversion',
        CT=>'transition',
        GA=>'transition',
        GC=>'transversion',
        GT=>'transversion',
        TA=>'transversion',
        TC=>'transition',
        TG=>'transversion'
    );

    warn "No var type for change: $_[0] $_[1]" unless $v_type{$_[0].$_[1]};
    return $v_type{$_[0].$_[1]};
}#FOLDEND


sub make_compression_from_file {#FOLDBEG
    my ($data_set, $algorithm, $seq_id) = @_;


    #Retrieve the encoding
    my %runlength_encoding;
    my $encoding_file = $encoding_dir . $data_set . '/' . 'runlength' 
            . '_' . $algorithm . '.txt';

    unless ( -e $encoding_file ) {
        print "Non-existent encoding - create the encoding first\n";
        return;
    }

    open my $INFILE, '<', $encoding_file ||
        die "Could not open encoding file";
    while (my $line = <$INFILE>) {
        chomp $line;
        my ($r, $e) = split /\s+/, $line;
        $runlength_encoding{$r} = $e;
    }
    close $INFILE;
            
    my %variant_encoding;
    $encoding_file = $encoding_dir . $data_set . '/' . 'variant' 
            . '_' . $algorithm . '.txt';
    open $INFILE, '<', $encoding_file ||
        die "Could not open encoding file";
    while (my $line = <$INFILE>) {
        chomp $line;
        my ($v, $e) = split /\s+/, $line;
        $variant_encoding{$v} = $e;
    }
    close $INFILE;


    # Retrieve the seq variants that are to be encoded
    my $data_file = $data_dir . $data_set . '/variants.fasta';
    my $variants_ref = get_sequence_variants($data_file, $seq_id);

 
    #Compute the output and their sizes - uncompressed and compressed
    my $txt_file = '/tmp/'.$seq_id.'__'.$data_set.'.txt';
    open my $TXT, '>', $txt_file 
        || die 'Could not open file';
    my $bit_file = '/tmp/'.$seq_id.'__'.$data_set.'__'.$algorithm.'.bit';
    open my $BIT, '>', $bit_file
        || die 'Could not open file';

    my $uncompressed_size = 0;
    my $compressed_size = 0;
    my $current_position = 0;
    for ( @{$variants_ref} ) {

        my $txt_entry = $_->[0] . $_->[1] . "\n";
        $uncompressed_size += (8 * length($txt_entry) );
        print $TXT $txt_entry;

        my $runlength = $_->[0] - $current_position;
        my $bit_entry = $runlength_encoding{$runlength} . $variant_encoding{$_->[1]};
        $compressed_size += length($bit_entry);
        print $BIT $bit_entry;
        $current_position = $_->[0];
    }

    close $TXT;
    close $BIT;


    #Give some output
    print "Sequence: $seq_id - data_set: $data_set - Algorithm: $algorithm\n";
    print "Data file: $txt_file     Compression file: $bit_file\n";
    print "Uncompressed size:\n";
    printf "%.2f bits\n", $uncompressed_size;
    printf "%.2f bytes\n", $uncompressed_size / 8;
    printf "%.2f k-bytes\n", $uncompressed_size / (8 * 1024);
    print "Compressed size:\n";
    printf "%.2f bits\n", $compressed_size;
    printf "%.2f bytes\n", $compressed_size / 8;
    printf "%.2f k-bytes\n", $compressed_size / (8 * 1024);

}#FOLDEND


sub make_compression_hapmap {#FOLDBEG
    my ($chromosome, $algorithm, $donor_id) = @_;

    #Get a database connection
    my $dbh = get_dbh($chromosome);


    #Retrieve the encoding
    my %runlength_encoding;
    my $encoding_file = $encoding_dir . 'hapmap/chr'. $chromosome . '_' . 'runlength' 
            . '_' . $algorithm . '.txt';

    unless ( -e $encoding_file ) {
        print "Non-existent encoding - create the encoding first\n";
        return;
    }

    open my $INFILE, '<', $encoding_file ||
        die "Could not open encoding file";
    while (my $line = <$INFILE>) {
        chomp $line;
        my ($r, $e) = split /\s+/, $line;
        $runlength_encoding{$r} = $e;
    }
    close $INFILE;
            
    my %variant_encoding;
    $encoding_file = $encoding_dir . 'hapmap/chr'. $chromosome . '_' . 'variant' 
            . '_' . $algorithm . '.txt';
    open $INFILE, '<', $encoding_file ||
        die "Could not open encoding file";
    while (my $line = <$INFILE>) {
        chomp $line;
        my ($v, $e) = split /\s+/, $line;
        $variant_encoding{$v} = $e;
    }
    close $INFILE;


    #Query the database for the donor identifier
    my ($identifier) = $dbh->selectrow_array( qq{
        SELECT code 
        FROM donor
        WHERE
            id = ?
        }, 
        {}, 
        ( $donor_id ) );
    

    #Query the database for the variant data
    my $row_ref = $dbh->selectall_arrayref( qq{
        SELECT position, variant
        FROM donor, variant, donor_variant
        WHERE
            donor.id = donor_variant.donor_id AND
            variant.id = donor_variant.variant_id AND
            donor.id = ?
        ORDER BY position ASC},
        {},
        ( $donor_id ) );


    #Compute the output and their sizes - uncompressed and compressed
    my $txt_file = '/tmp/'.$identifier.'__hapmap_chr'.$chromosome.'.txt';
    open my $TXT, '>', $txt_file 
        || die 'Could not open file';
    my $bit_file = '/tmp/'.$identifier.'__hapmap_chr'.$chromosome.'__'.$algorithm.'.bit';
    open my $BIT, '>', $bit_file
        || die 'Could not open file';

    my $uncompressed_size = 0;
    my $compressed_size = 0;
    my $current_position = 0;
    for ( @{$row_ref} ) {

        my $txt_entry = $_->[0] . $_->[1] . "\n";
        $uncompressed_size += (8 * length($txt_entry) );
        print $TXT $txt_entry;

        my $runlength = $_->[0] - $current_position;
        my $bit_entry = $runlength_encoding{$runlength} . $variant_encoding{$_->[1]};
        $compressed_size += length($bit_entry);
        print $BIT $bit_entry;
        $current_position = $_->[0];
    }

    close $TXT;
    close $BIT;


    #Give some output
    print "Sequence: $donor_id - Chromosome: $chromosome - Algorithm: $algorithm\n";
    print "Data file: $txt_file     Compression file: $bit_file\n";
    print "Uncompressed size:\n";
    printf "%.2f bits\n", $uncompressed_size;
    printf "%.2f bytes\n", $uncompressed_size / 8;
    printf "%.2f k-bytes\n", $uncompressed_size / (8 * 1024);
    print "Compressed size:\n";
    printf "%.2f bits\n", $compressed_size;
    printf "%.2f bytes\n", $compressed_size / 8;
    printf "%.2f k-bytes\n", $compressed_size / (8 * 1024);

}#FOLDEND


sub make_encoding_elias_gamma {#FOLDBEG

    my $key_counts = shift;  #a hash reference
    
    
    my @sorted_keys = sort {$key_counts->{$b} <=> $key_counts->{$a}} 
        keys %{$key_counts};
    
    my @encoding;
    for (my $i=0; $i < @sorted_keys; $i++) {
        $encoding[$i] = $sorted_keys[$i] . ',' . encode_elias_gamma($i);
    }
        
    return \@encoding;


}#FOLDEND


sub make_encoding_golomb {#FOLDBEG
    my ($key_counts, $key_freqs) = @_;  #a hash reference
    
    #Get the inverse of the expected value
    my @sorted_keys = sort {$key_counts->{$b} <=> $key_counts->{$a}} 
        keys %{$key_counts};

    my $exp_value = 0;
    for (my $i=0; $i < @sorted_keys; $i++) {
        $exp_value += ($i + 1) * $key_freqs->{$sorted_keys[$i]};
    }

    my $p = 1/$exp_value;

    #Compute the m-value
    my $m = int( log($p) / log(1 - $p));
    $m = 1 if $m < 1;

    my @encoding;
    for (my $i=0; $i < @sorted_keys; $i++) {
        $encoding[$i] = $sorted_keys[$i] . ',' . encode_golomb($i, $m);
    }
        
    return \@encoding;

}#FOLDEND


sub make_encoding_huffman {#FOLDBEG
    my ($key_counts) = @_;

    print "Creating Huffman encoding\n";
   
    #Get an descending ordered list of the keys
    my @sorted_keys = sort { $key_counts->{$b} <=> $key_counts->{$a} } 
        keys %{$key_counts};

    #Set an index value for the keys that will need an encoding
    my $end_index = $#sorted_keys;

    #Initialze some arrays that will keep track of node values.  Note: Using
    #an array or hash for each node uses too much memory for large trees.
    my @counts;
    my @parents;
    my @bits;
    for ( @sorted_keys ) {
        push @counts, $key_counts->{$_};
        push @parents, undef;
        push @bits, undef;
    }
    
    #finish the tree
    my $min_index1;
    my $min_index2;
    while (1) {
        #Extract the two minimum nodes
        for ( my $i = 0; $i < @parents; $i++ ){
            next if $parents[$i];
            
            if ( !defined($min_index1) ) {
                $min_index1 = $i;
                next;
            } 
            
            if ( !defined($min_index2) ){
                if ($counts[$i] < $counts[$min_index1] ) {
                    $min_index2 = $min_index1;
                    $min_index1 = $i;
                }
                else {
                    $min_index2 = $i;
                }
                next;
            }
            
            if ($counts[$i] < $counts[$min_index1]) {
                $min_index2 = $min_index1;
                $min_index1 = $i;
            }
            elsif ($counts[$i] < $counts[$min_index2]) {
                $min_index2 = $i;
            }
        }
        
        #Process the two minimum nodes or terminate
        if ( defined $min_index1 && defined $min_index2 ) {
            my $combined_count = $counts[$min_index1]
                + $counts[$min_index2];
           
            push @counts, $combined_count;
            push @parents, undef;
            push @bits, undef;
            
            $parents[$min_index1] = $#parents;
            $bits[$min_index1] = 'O';
            $parents[$min_index2] = $#parents;
            $bits[$min_index2] = 'l';
            
            $min_index1 = undef;
            $min_index2 = undef;
        }
        else {
            last;
        }
    }
    
    #extract the encodings
    my @encoding;
    for (my $i = 0; $i <= $end_index; $i++){
        my $code_string;
        my $p = $parents[$i];
        my $b = $bits[$i];

        while (1) {
            last unless $b;
            no warnings;
      
            $code_string .= $b;
            $b = $bits[$p];
            $p = $parents[$p];
        }
        
        push @encoding, $sorted_keys[$i] . ',' . reverse $code_string;
        $code_string = undef;
    }
    
    return \@encoding;
}#FOLDEND


sub make_encoding_unary {#FOLDBEG
    my $key_counts = shift;  #a hash reference
    
    
    my @sorted_keys = sort {$key_counts->{$b} <=> $key_counts->{$a}} 
        keys %{$key_counts};
    
    my @encoding;
    for (my $i=0; $i < @sorted_keys; $i++) {
        $encoding[$i] = $sorted_keys[$i] . ',' . encode_unary($i);
    }
        
    return \@encoding;
}#FOLDEND


sub menu_algorithm_analysis {#FOLDBEG

    #We are analyzing either hapmap, mtdna or your_data data.  Unless it is 
    #hapmap then the chromosome will be an empty string.
    my ($data_set, $chromosome) = @_;

    my $response;

    # Remove the unary option from the menu 'Unary',
   
    while ($response = prompt 'Which algorithm?',
        -menu=>
            [
                'Position-Variant',
                'Elias-Gamma',
                'Golomb',
                'Huffman',
                'Compression Factors',
	            'Go Back',
	        ])
     {
        my $results_file;
        if ($response eq 'Go Back') {
            return;
        }   
        elsif ( $response eq 'Compression Factors' ) {
            $results_file = analyze_compression_factors( $data_set, $chromosome );
            print "Analysis saved in $results_file\n";
        }
        elsif ( $response eq 'Position-Variant' ) {
            if ( $chromosome ) { 
                $results_file = analyze_encoding_hapmap_pv();
                print "Analysis saved in $results_file\n";
            }
            else {
                $results_file = analyze_encoding_pv($data_set);
                print "Analysis saved in $results_file\n";
            }
        }
        else {
            $results_file = analyze_encoding( $data_set, $chromosome, $response);
            print "Analysis saved in $results_file\n";
        }
    }
}#FOLDEND


sub menu_algorithm_compression {#FOLDBEG

    #We are analyzing hapmap, mtdna, or your_data data.
    #Hapmap data compressions send the chromosome number as the data_type
    #value.  mtDNA compressions send the reference sequence to use (mtdna_rcrs
    #or mtdna_consensus) as the data_type value.  Custom data compressions
    #set the data_type to 'your_data';
    my ( $data_type ) = @_;

    my $response;
   
    # Remove the unary option from the menu 'Unary',
    while ($response = prompt 'Which algorithm?',
        -menu=>
            [
                'Elias-Gamma',
                'Golomb',
                'Huffman',
	            'Go Back',
	        ])
     {

        if ($response eq 'Go Back') {
            return;
        }
        else {
            menu_donor_compression( $data_type, $response );
        }
    }
}#FOLDEND


sub menu_algorithm_encoding {#FOLDBEG
    my ($data_set) = @_;

    my $runlength_counts;
    my $variant_counts;

    # Unary option removed from the menu - 'Unary',
    while (my $algorithm = prompt 'Algorithm to use:',
        -menu=>
            [
                'Elias-Gamma',
                'Golomb',
                'Huffman',
                'Go Back',
            ])
    {

        $runlength_counts = get_key_data($data_set, 'runlength', 'counts')
            unless $runlength_counts;
        $variant_counts = get_key_data($data_set, 'variant', 'counts')
            unless $variant_counts;

        my $runlength_encoding;
        my $variant_encoding;
        if ($algorithm eq 'Go Back') {
            last;
        }
        elsif ($algorithm eq 'Huffman') {
            $runlength_encoding = make_encoding_huffman($runlength_counts);
            $variant_encoding = make_encoding_huffman($variant_counts);
        }
        elsif ($algorithm eq 'Golomb') {
            my $runlength_freqs = get_key_data($data_set, 'runlength', 'freqs');
            my $variant_freqs = get_key_data($data_set, 'variant', 'freqs');
            $runlength_encoding = make_encoding_golomb($runlength_counts, $runlength_freqs);
            $variant_encoding = make_encoding_golomb($variant_counts, $variant_freqs);
        }
        elsif ($algorithm eq 'Elias-Gamma') {
            $runlength_encoding = make_encoding_elias_gamma($runlength_counts);
            $variant_encoding = make_encoding_elias_gamma($variant_counts);
        }
        elsif ($algorithm eq 'Unary') {
            $runlength_encoding = make_encoding_unary($runlength_counts);
            $variant_encoding = make_encoding_unary($variant_counts);
        }

        store_encoding($data_set, $algorithm, $runlength_encoding, 'runlength');
        store_encoding($data_set, $algorithm, $variant_encoding, 'variant');

        return;

    }
}#FOLDEND


sub menu_analysis {#FOLDBEG
    my $response;
   
    while ($response = prompt 'Which dataset?',
        -menu=>
            [
                'HapMap',
	            'mtDNA Consensus',
                'mtDNA rCRS',
                'Your Data',
	            'Go Back',
	        ])
     {

        if ($response eq 'HapMap') {
            menu_chromosome_analysis();
        }
        elsif ($response eq 'mtDNA Consensus') {
            menu_algorithm_analysis('mtdna_consensus', '');
        }   
        elsif ($response eq 'mtDNA rCRS') {
            menu_algorithm_analysis('mtdna_rcrs', '');
        }   
        elsif ($response eq 'Your Data') {
            menu_algorithm_analysis('your_data', '');
        }   
        elsif ($response eq 'Go Back') {
            return;
        }   
    }
}#FOLDEND


sub menu_compression {#FOLDBEG
    my $data_set;
    my $algorithm;
    while ($data_set = prompt 'Type of compression:',
        -menu=>
            [ 
	            'HapMap',
	            'mtDNA Consensus',
	            'mtDNA rCRS',
	            'Your Data',
	            'Go Back',
            ])
    {

        if ($data_set eq 'HapMap') {
            # Remove the chromosome option from the menu
            #menu_chromosome_compression();
            menu_algorithm_compression('1');
        }
        elsif ($data_set eq 'mtDNA Consensus') {
            menu_algorithm_compression('mtdna_consensus');
        }
        elsif ($data_set eq 'mtDNA rCRS') {
            menu_algorithm_compression('mtdna_rcrs');
        }
        elsif ($data_set eq 'Your Data') {
            menu_algorithm_compression('your_data');
        }
        elsif ($data_set eq 'Go Back') {
            return;
        }
    }
}#FOLDEND


sub menu_chromosome_analysis {#FOLDBEG
    my $chromosome;
    while ($chromosome = prompt 
        'Which chromosome? ( 1-22, X, Y ) or q to return: ') {
        
        if ( $chromosome =~ /^q$/i ){
            return;
        }

        unless ( $chromosome =~ /^[1-2][1-2]?$/ or $chromosome =~ /^X$/i 
                or $chromosome =~ /^Y$/i ) {
            next;
        }

        menu_algorithm_analysis('hapmap', $chromosome);
    }

}#FOLDEND


sub menu_decompression {#FOLDBEG

    while (my $file_name = prompt 
        'Name of file to decompress or q to return:') {
        
        if ( $file_name =~ /^q$/i ){
            return;
        }

        unless ( -e $file_name ) {
            print "Invalid file name\n";
            next;
        }

        decompress($file_name);
    }
}#FOLDEND


sub menu_donor_compression {#FOLDBEG

    my ($data_type, $algorithm) = @_;

    while (my $seq_id = prompt 
        "Enter a ID value from the dataset or q to return: ") {
        
        if ( $seq_id =~ /^q$/i ){
            return;
        }

        unless ( $seq_id =~ /^\d{1,4}$/ ) {
            print "Invalid donor ID\n";
            next;
        }

        unless ( $seq_id > 0 ) {
            print "Invalid donor ID\n";
            next;
        }

        if ( $data_type eq 'mtdna_rcrs' or $data_type eq 'mtdna_consensus' ) {
            make_compression_from_file($data_type, $algorithm, $seq_id);
        }
        elsif ( $data_type eq 'your_data' ) {
            make_compression_from_file($data_type, $algorithm, $seq_id);
        }
        else {
            make_compression_hapmap($data_type, $algorithm, $seq_id);
        }
    }

}#FOLDEND


sub menu_encoding {#FOLDBEG
    my $data_set;
    my $algorithm;
    while ($data_set = prompt 'Which data?',
        -menu=>
            [ 
	            'HapMap',
	            'mtDNA Consensus',
	            'mtDNA rCRS',
                'Your Data',
	            'Go Back',
            ])
    {

        if ($data_set eq 'HapMap') {
            menu_algorithm_encoding(1);

            # The line below can be used to give a choice of chromosome
            #menu_chromosome_encoding();
        }
        elsif ($data_set eq 'mtDNA Consensus') {
            menu_algorithm_encoding('mtdna_consensus');
        }
        elsif ($data_set eq 'mtDNA rCRS') {
            menu_algorithm_encoding('mtdna_rcrs');
        }
        elsif ($data_set eq 'Your Data') {
            menu_algorithm_encoding('your_data');
        }
        elsif ($data_set eq 'Go Back') {
            return;
        }
    }
}#FOLDEND


sub menu_get_data {#FOLDBEG
    while (my $response = prompt 'What data do you want?',
        -menu=>
            [
	            'Download HapMap Genotype Data',
                'HapMap Genotype Data',
                'Pos-Variant HapMap Data Analysis',
                'HapMap Frequencies',
	            'mtDNA rCRS Data',
	            'mtDNA Consensus Data',
	            'Go Back',
	        ])
    {  
        if ( $response eq 'Download HapMap Genotype Data' ) {
            #print "Downloading files - Go have coffee.\n";
            download_hapmap_files();
        }
        elsif ( $response eq 'HapMap Genotype Data' ) {
            extract_hapmap_genotype_data();
        }
        elsif ( $response eq 'HapMap Frequencies' ) {
            compute_hapmap_frequencies();
        }
        elsif ( $response eq 'mtDNA rCRS Data' ) {
            get_mtDNA_variants('rCRS');
        }

        elsif ( $response eq 'mtDNA Consensus Data' ) {
            get_mtDNA_variants('consensus');
        }

        else {
            return;
        }
    }
}#FOLDEND


sub menu_chromosome_compression {#FOLDBEG
    my $chromosome;
    while ($chromosome = prompt 
        'Which chromosome? ( 1-22, X, Y ) or q to return: ') {
        
        if ( $chromosome =~ /^q$/i ){
            return;
        }

        unless ( $chromosome =~ /^[1-2][1-2]?$/ or $chromosome =~ /^X$/i 
                or $chromosome =~ /^Y$/i ) {
            next;
        }

        menu_algorithm_compression($chromosome);
    }

}#FOLDEND


sub menu_chromosome_encoding {#FOLDBEG
    my $chromosome;
    while ($chromosome = prompt 
        'Which chromosome? ( 1-22, X, Y ) or q to return: ') {
        
        if ( $chromosome =~ /^q$/i ){
            return;
        }

        unless ( $chromosome =~ /^[1-2][1-2]?$/ or $chromosome =~ /^X$/i 
                or $chromosome =~ /^Y$/i ) {
            next;
        }

        menu_algorithm_encoding($chromosome);
    }

}#FOLDEND


sub menu_top {#FOLDBEG
    my $response;

    # Prompt menus cause warnings about variable constants.
    # Add 'Get Data' to the menu below to get the options for downloading data
    while ($response = prompt "\nWhat to do?",
        -menu=>
            [
	            'Create Encoding',
	            'Analyze Encoding',
	            'Compress',
	            'Decompress',
	            'Quit',
	        ]){

                if ($response eq 'Create Encoding') {
                    menu_encoding();
                }
                elsif ($response eq 'Analyze Encoding') {
                    menu_analysis();
                }
                elsif ($response eq 'Compress') {
                    menu_compression();
                }
                elsif ($response eq 'Decompress') {
                    menu_decompression();
                }
                elsif ($response eq 'Get Data') {
                    menu_get_data();
                }
                elsif ($response eq 'Quit') {
                    exit;
                }
        }
}#FOLDEND


sub make_consensus_adjustments {#FOLDBEG
    #Determines the consensus haplotypes, and removes
    #those variants that match the consensus haplotype

    my ( $chromosome, $data_grid ) = @_;

    print "Making consensus adjustments\n";

    #Open the file that will contain the consensus entries
    my $consensus_file = $reference_dir . 'chr'. $chromosome . '_hapmap_consensus.txt';
    open my $CONSENSUS, '>>', $consensus_file
        || die "Could not open file: $consensus_file";


    for ( sort { $a <=> $b } keys %{ $data_grid } ) {

        #Get counts for all the haplotypes at that position
        my @haplotypes = split ',', $data_grid->{ $_ };
        my %hap_counts;
        map { $hap_counts{$_}++ if $_ } @haplotypes;

        #Determine the consensus haplotype
        my $consensus_haplotype;
        my $consensus_count = 0;
        for ( keys %hap_counts ) {
            if ( $hap_counts{ $_ } > $consensus_count ) {
                $consensus_count = $hap_counts{ $_ };
                $consensus_haplotype = $_;
            }
        }

        #Make an entry in the consensus file
        print $CONSENSUS $_, $consensus_haplotype, "\n";

        #Remove values equal to the consensus;
        for ( @haplotypes ) {
           $_ = '0' if $_ eq $consensus_haplotype;
        }
        $data_grid->{ $_ } = join ',', @haplotypes;
    }

    close $CONSENSUS;

}#FOLDEND


sub store_encoding {#FOLDBEG
    my ($data_set, $algorithm, $encoding, $type) = @_;

    #print "Storing $algorithm encoding for $type\n";
    
    #Open the file that will contain the encoding
    my $encoding_file;
    
    if ( $data_set eq 'mtdna_consensus' ) {
        $encoding_file = $encoding_dir . 'mtdna_consensus/' . $type . '_' 
            . $algorithm . '.txt';
    }
    elsif ( $data_set eq 'mtdna_rcrs' ) {
        $encoding_file = $encoding_dir . 'mtdna_rcrs/' . $type . '_' . $algorithm . '.txt';
    }
    elsif ( $data_set eq 'your_data' ) {
        $encoding_file = $encoding_dir . 'your_data/' . $type . '_' . $algorithm . '.txt';
    }
    else {
        $encoding_file = $encoding_dir . 'hapmap/chr'. $data_set . '_' . $type 
            . '_' . $algorithm . '.txt';
    }

    open my $ENCODING, '>', $encoding_file
        || die "Could not open file: $encoding_file";

    for ( @{$encoding} ) {
        my ($k, $c) = split ',', $_;
        if ($type eq 'runlength') {
            printf $ENCODING "%-15s%-s\n", $k, $c;
        }
        else {
            printf $ENCODING "%-20s%-s\n", $k, $c;
        }
    }

    close $ENCODING;
}#FOLDEND


sub store_hapmap_data {#FOLDBEG
    my ( $chromosome, $population, $data_grid, $donor_list ) = @_;

    print "Storing haplotype data - chromosome $chromosome\n";

	#Create new sqlite databases
    system './sqlite_ddl.sh';
    

    #Get a connection
	my $dbh = get_dbh( $chromosome );

    my $donor_list_stm = $dbh->prepare( qq{
        SELECT code, id 
        FROM donor
    } );

    my $donor_insert_stm = $dbh->prepare( qq{ 
        INSERT INTO donor ( id, population, code ) VALUES (?,?,?)
    } );

    my $variant_insert_stm = $dbh->prepare( qq{ 
        INSERT INTO variant (id, position, variant, chromosome) 
        VALUES (?,?,?,?)
    } );

    my $donor_variant_insert_stm = $dbh->prepare( qq{ 
        INSERT INTO donor_variant (donor_id, variant_id) 
        VALUES (?,?)
    } );


    #Cache the donor ID values to make it a bit faster
    my ($d_id_counter) = $dbh->selectrow_array ( qq{
        SELECT MAX(id) FROM donor 
    });
    $d_id_counter = $d_id_counter ? ++$d_id_counter : 1; 
    $donor_list_stm->execute();
    my $row_ref = $donor_list_stm->fetchall_arrayref();
    my %d_id_cache;
    map { $d_id_cache{$_->[0]} = $_->[1] } @{$row_ref};


    #Cache the variant ID values to make it a bit faster
    my ($v_id_counter) = $dbh->selectrow_array ( qq{
        SELECT MAX(id) FROM variant
    });
    $v_id_counter = $v_id_counter ? ++$v_id_counter : 1; 
    my %v_id_cache;

    #Make any needed donor entries
    $dbh->do( qq{ BEGIN TRANSACTION } );
    foreach my $d ( @{$donor_list} ) {

        #make the donor entry
        if ( $d_id_cache{$d} ){
            next;
        }
        else {
            $donor_insert_stm->execute( $d_id_counter, $population, $d );
            $d_id_cache{$d} = $d_id_counter;
            $d_id_counter++;
        }
    }
    $dbh->do( qq{ END TRANSACTION } );

    #Make the variant entries
    $dbh->do( qq{ BEGIN TRANSACTION } );
    for ( keys %{$data_grid} ) {
        my $p = $_;
        my @haplotypes = split ',', $data_grid->{$_};

        for (my $i = 0; $i < @haplotypes; $i++) {

            #Don't enter variants which match the consensus - zero valued
            next unless $haplotypes[$i];

            #Make any needed variant entry
            my $v = $haplotypes[$i]; 
            my $v_id;
            if ( $v_id_cache{$v.$p} ) {
                #the variant has previously been entered
                $v_id = $v_id_cache{$v.$p};
            }
            else {
                #a new variant, generate the ID and make an entry
                $v_id = $v_id_counter;
                $v_id_cache{$v.$p} = $v_id;
                $v_id_counter++;
                $variant_insert_stm->execute( $v_id, $p, $v, $chromosome );
            }


            #Make the donor_variant entries
            $donor_variant_insert_stm->execute( $d_id_cache{$donor_list->[$i]}, $v_id );

        }
    }
    $dbh->do( qq{ END TRANSACTION } );
}#FOLDEND
